/*
	ps2txt.c		(c) Masahiro Kajiura
	(Ver. 0.4.1)		(masahiro.kajiura@toshiba.co.jp)

	* This program is called by ghostscript(gs) in shell-command 'ps2text'.
*/

#include <stdio.h>

#define SHOWPAGESTR	"="
#define DIVIDER		' '
#define INDENTUNIT	"  "
#define TRUE		(1==1)
#define FALSE		(!TRUE)

typedef struct _substr {
    double		x1, x2, y;
    unsigned char	*str;
    struct _substr	*next;
} SUBSTR;

typedef struct _headloc {
    double		x;
    int			num;
    struct _headloc	*next;
} HEADLOC;

typedef struct _line {
    SUBSTR		*left, *right;
    HEADLOC		*headloc;
    struct _line	*next;
    double		y;
} LINE;

LINE	*top, *bottom;
int	linenum;
HEADLOC	*topheadloc;
double	curspace = -1.0;

main()
{
    int	firstpage = TRUE;

    while( getpagedata() ) {
	if( firstpage )
	    firstpage = FALSE;
	else
	    printf( "\014\n" );		/* ^L */
	printout();
	freeall();
    }
}

getpagedata()
{
    unsigned char	buf[BUFSIZ];
    SUBSTR		*ptr;
    SUBSTR		*getsubstr();
    int			cont = FALSE;
    double		atof();

    top = bottom = NULL;
    topheadloc = NULL;
    linenum = 0;

    while( gets( buf ) ) {
	cont = TRUE;
	if( strcmp( buf, SHOWPAGESTR ) == 0 )
	    break;
	if( *buf == 'S' ) {
	    curspace = atof( buf + 2 );
	    continue;
	}
	ptr = getsubstr( buf );
	if( !ptr )
	    continue;
	appendsubstr( ptr );
    }
    return cont;
}

appendsubstr( sptr )
SUBSTR	*sptr;
{
    LINE	*ptr, *tmp;
    LINE	*mallocline();

    if( !bottom ) {
	top = bottom = mallocline();
	top->y = sptr->y;
	top->left = top->right = sptr;
	return;
    }
    if( bottom->y == sptr->y ) {
	appendsubstrtoline( bottom, sptr );
	return;
    }
    if( bottom->y < sptr->y ) {
	bottom->next = mallocline();
	bottom = bottom->next;
	bottom->left = bottom->right = sptr;
	bottom->y = sptr->y;
	return;
    }
    if( top->y > sptr->y ) {
	ptr = mallocline();
	ptr->next = top;
	top = ptr;
	top->y = sptr->y;
	top->left = top->right = sptr;
	return;
    }
    for( ptr = top; ptr->next; ptr = ptr->next ) {
	if( ptr->y == sptr->y ) {
	    appendsubstrtoline( ptr, sptr );
	    return;
	}
	if( ptr->next->y > sptr->y ) {
	    tmp = mallocline();
	    tmp->next = ptr->next;
	    ptr->next = tmp;
	    tmp->y = sptr->y;
	    tmp->left = tmp->right = sptr;
	    return;
	}
    }
}

appendsubstrtoline( lptr, sptr )
LINE	*lptr;
SUBSTR	*sptr;
{
    SUBSTR	*ptr;

    if( lptr->right->x1 <= sptr->x1 ) {
	lptr->right->next = sptr;
	lptr->right = sptr;
    } else if( lptr->left->x1 > sptr->x1 ) {
	sptr->next = lptr->left;
	lptr->left = sptr;
    } else {
	for( ptr = lptr->left; ptr->next; ptr = ptr->next ) {
	    if( ptr->next->x1 > sptr->x1 ) {
		sptr->next = ptr->next;
		ptr->next = sptr;
		return;
	    }
	}
    }
}

LINE *mallocline()
{
    LINE *ptr;

    linenum++;
    ptr = (LINE *)malloc( sizeof( LINE ) );
    if( !ptr ) {
	fprintf( stderr, "cannot malloc for struct line\n" );
	exit( 1 );
    }
    ptr->next = NULL;

    return ptr;
}

SUBSTR *getsubstr( data )
unsigned char	*data;
{
    double		x1, y1, x2, y2;
    unsigned char	*strhead;
    int			i;
    SUBSTR		*ptr;
    SUBSTR		*mallocsubstr();

    sscanf( data, "%lf %lf %lf %lf", &x1, &y1, &x2, &y2 );
    strhead = data;
    for( i = 0; i < 4; ) {
	while( *strhead++ != DIVIDER )
	    ;
	i++;
    }
    if( y1 != y2 ) {
	fprintf( stderr, "vertical string \"%s\"\n", strhead );
	return NULL;
    }
    if( x1 > x2 ) {
	fprintf( stderr, "reverse direction string \"%s\"\n", strhead );
	return NULL;
    }
    ptr = mallocsubstr( strlen( strhead ) );
    ptr->x1 = x1;
    ptr->x2 = x2;
    ptr->y  = y1;
    strcpy( ptr->str, strhead );

    return ptr;
}

SUBSTR *mallocsubstr( len )
int	len;
{
    SUBSTR *ptr;

    ptr = (SUBSTR *)malloc( sizeof( SUBSTR ) );
    if( !ptr ) {
	fprintf( stderr, "cannot malloc for substr\n" );
	exit( 1 );
    }

    ptr->str = (unsigned char *)malloc( len + 1 );
    if( !(ptr->str) ) {
	fprintf( stderr, "cannot malloc for str in substr\n" );
	exit( 1 );
    }

    ptr->next = NULL;
    return ptr;
}

printout()
{
    LINE	*lptr;
    SUBSTR	*sptr;
    int		i, spacenum;
    double	spacewidth;

    sortheadloc();
    for( lptr = top; lptr; lptr = lptr->next ) {
	for( i = lptr->headloc->num; i--; )
	    printf( "%s", INDENTUNIT );
	for( sptr = lptr->left; sptr; sptr = sptr->next ) {
	    printf( "%s", sptr->str );
	    if( sptr->next ) {
		if( curspace > 0.0 )
		    spacewidth = curspace * 0.9;	/* <= sensitive! */
		else
		    spacewidth = 0.5 *			/* <= sensitive! */
			(sptr->x2 - sptr->x1 + sptr->next->x2 - sptr->next->x1)
			/ ( strlen(sptr->str) + strlen(sptr->next->str) );
		if( sptr->next->x1 - sptr->x2 > spacewidth ) {
		    spacenum = ( spacewidth == 0 ) ? 1
			: ( ( sptr->next->x1 - sptr->x2 ) / spacewidth );
		    spacenum = spacenum < 1 ? 1 : spacenum;
		    for( i =  spacenum; i--; )
			putchar( ' ' );
		}
	    }
	}
	putchar( '\n' );
    }
}

freeall()
{
    LINE	*lptr, *lnext;
    SUBSTR	*sptr, *snext;
    HEADLOC	*hptr, *hnext;

    for( lptr = top; lptr; lptr = lnext ) {
	lnext = lptr->next;
	for( sptr = lptr->left; sptr; sptr = snext ) {
	    snext = sptr->next;
	    free( sptr->str );
	    free( sptr );
	}
	free( lptr );
    }
    for( hptr = topheadloc; hptr; hptr = hnext ) {
	hnext = hptr->next;
	free( hptr );
    }
}

sortheadloc()
{
    LINE	*lptr;
    HEADLOC	*hptr, *htmp;
    HEADLOC	*mallocheadloc();
    int		i = 0;

    for( lptr = top; lptr; lptr = lptr->next ) {
	if( !topheadloc ) {
	    topheadloc = mallocheadloc();
	    topheadloc->x = lptr->left->x1;
	    lptr->headloc = topheadloc;
	    continue;
	}
	if( topheadloc->x > lptr->left->x1 ) {
	    htmp = mallocheadloc();
	    htmp->x = lptr->left->x1;
	    htmp->next = topheadloc;
	    lptr->headloc = topheadloc = htmp;
	    continue;
	}
	for( hptr = topheadloc; hptr->next; hptr = hptr->next ) {
	    if( hptr->x == lptr->left->x1 ) {
		lptr->headloc = hptr;
		goto loopexit;
	    }
	    if( hptr->next->x > lptr->left->x1 ) {
		htmp = mallocheadloc();
		htmp->x = lptr->left->x1;
		htmp->next = hptr->next;
		hptr->next = lptr->headloc = htmp;
		goto loopexit;
	    }
	}
	if( hptr->x == lptr->left->x1 ) {
	    lptr->headloc = hptr;
	    continue;
	}
	if( hptr->x < lptr->left->x1 ) {
	    htmp = mallocheadloc();
	    htmp->x = lptr->left->x1;
	    hptr->next = htmp;
	    lptr->headloc = htmp;
	}
	loopexit:;
    }
    for( hptr = topheadloc; hptr; hptr = hptr->next ) {
	hptr->num = i++;
    }
}

HEADLOC *mallocheadloc()
{
    HEADLOC *ptr;

    ptr = (HEADLOC *)malloc( sizeof( HEADLOC ) );
    if( !ptr ) {
	fprintf( stderr, "cannot malloc for struct headloc\n" );
	exit( 1 );
    }
    ptr->next = NULL;

    return ptr;
}
